print("import: 'redis'")
import redis

print("import: 'redis.asyncio'")
import redis.asyncio

print("import: 'redis.commands'")
import redis.commands

print("import: 'redis.commands.bf'")
import redis.commands.bf

print("import: 'redis.commands.json'")
import redis.commands.json

print("import: 'redis.commands.search'")
import redis.commands.search

print("import: 'redis.commands.timeseries'")
import redis.commands.timeseries

print("import: 'redis.commands.graph'")
import redis.commands.graph

